
/*********************************************************\
*  http://www.vems.hu/
*	BOOSTCONTROL                                      *
*  (C) 2004 [Marcell Gal, cell at x-dsl.hu                *
**********************************************************/
#include "global.h"

#ifdef BOOSTCONTROL

#include "boostcontrol.h"

#include "eventqueue.h"
#include "helpers.h"
#include "adc.h"
#include "pid.h"
#include "spi.h"

#include "lbus.h"
#include "softpwm.c"
#include "injector.c"

#ifdef BOOSTGAUGE
struct softpwm_t boostgauge_pwm;

/**
* Fills engine.boost_ref_dc according to 2 dimension (RPM, MAP) simple lookup
* t8: an input argument, so same function can be used for relative and absolute
* lookup
*/
static void boostref_table_lookup(uint8_t t8)
{               
uint8_t tempref; //temp for refDC
#ifndef PS2KEYBOARD
	// what is this ???
	if ((config_ext.boostsw_conf&0x10)&&(boost_openloop)&&(boostsimple_configured))
	{
		struct search_table_t myspeed_idx;
		uint8_t uindex, lindex;

		search_table(ve.launch_rpmtable+8, 8, engine.speed, &myspeed_idx);

		uindex = myspeed_idx.uindex;
		lindex = myspeed_idx.lindex;
		tempref = linear_interp(ve.launch_rpmtable[8+lindex], ve.launch_rpmtable[8+uindex],
			boost.boost_ref_table[lindex], boost.boost_ref_table[uindex], engine.speed );

	}
	else
	{
#endif
	// simple openloop uses first line of boost_ref_table
	if ((boost_openloop)&&(boostsimple_configured))
	{	
		struct search_table_t myrpm_idx;
		uint8_t uindex, lindex;
	
		search_table(boost.boost_ref_rpm_range, sizeof(boost.boost_ref_rpm_range), engine.rpm, &myrpm_idx);
		
		uindex = myrpm_idx.uindex;
		lindex = myrpm_idx.lindex; 
#ifndef PS2KEYBOARD		
   
		 //dont use first 6 row in pwm table, use 7th row	for simple boost
		tempref = linear_interp(boost.boost_ref_rpm_range[lindex], boost.boost_ref_rpm_range[uindex],
			boost.boost_ref_table[6*IGNADV_SIZE_RPM+lindex], boost.boost_ref_table[6*IGNADV_SIZE_RPM+uindex], engine.rpm ); 

#else
		tempref = linear_interp(boost.boost_ref_rpm_range[lindex], boost.boost_ref_rpm_range[uindex],
			boost.boost_ref_table[lindex], boost.boost_ref_table[uindex], engine.rpm );
#endif
	}
	// simple "closedloop" uses full boost_ref_table, does others also (rel, abs)
	else
	{
		struct search_table_t myrpm_idx, mykpa_idx;
		uint16_t rpm_weightu, kpa_weightu;
      	 
		search_table(boost.boost_ref_kpa_range, sizeof(boost.boost_ref_kpa_range), t8,         &mykpa_idx);
#ifndef PS2KEYBOARD		
		if ((boost_PWM_gear_depending)&&(mykpa_idx.lindex<6)) {       // first 6 rox are used for gear depending boost, so for that row value is determinated by config.boost_conf bit 3 and 2(ref possition)  0%, 33% 66% and 100%
		  tempref =  ((config.boost_conf & 0x0C)>>2)*85;     //return ref posstion
		   }
		   else{                                                         // return value from usefull part of PWM table(12x6) for water,N2O or..
		   search_table(boost.boost_ref_rpm_range, sizeof(boost.boost_ref_rpm_range), engine.rpm, &myrpm_idx);
		   rpm_weightu = calc_weightu( engine.rpm, boost.boost_ref_rpm_range, &myrpm_idx );
		   kpa_weightu = calc_weightu( t8        , boost.boost_ref_kpa_range, &mykpa_idx );

		   tempref = lookup_and_weight( boost.boost_ref_table, &myrpm_idx, &mykpa_idx, rpm_weightu, kpa_weightu, IGNADV_SIZE_RPM );
		   }
#else
		   search_table(boost.boost_ref_rpm_range, sizeof(boost.boost_ref_rpm_range), engine.rpm, &myrpm_idx);
		   rpm_weightu = calc_weightu( engine.rpm, boost.boost_ref_rpm_range, &myrpm_idx );
		   kpa_weightu = calc_weightu( t8        , boost.boost_ref_kpa_range, &mykpa_idx );

		   tempref = lookup_and_weight( boost.boost_ref_table, &myrpm_idx, &mykpa_idx, rpm_weightu, kpa_weightu, IGNADV_SIZE_RPM );
#endif

	}      
   
#ifndef PS2KEYBOARD
   }               
#endif
       
if (boost_EGT_MAT_refDC_reducing) {            // reducing boostDC ref when egt or iat is to big

	uint16_t egt16, t16; // 
	uint8_t t8, threshold;	  
	
	if( (t8=engine.mat) > (threshold = config.boost_matdec) ){
		t8 -= threshold; 
		t8 <<= 1;   // +10C removing 7% of DC  
		tempref = tempref > t8 ? tempref-t8 : 0;
	}

	egt16 = engine.egt;
	t8 = egt16 >> 8; // rail above 2040C, not overflow (never happen)
	t8 = t8 > 7 ? 255 : egt16>>3; // going to [8C]
	if( t8 > (threshold = config.boost_egtdec) ){
		t8 -= threshold;
		t8 <<=2 ; // +40C = +5 => -7.8% of DC
		tempref = tempref > t8 ? tempref-t8 : 0;
	}
	
	
	 
	 
	}


	engine.boost_ref_dc = t8; //return refDC 

	
	
}

/** 1 dimension (RPM) lookup
*/
static void boostrpm_lookup(void)
{
	struct search_table_t myrpm_idx;
	uint16_t rpm_weightu, weighted;
	uint8_t b1, b2;
	search_table(boost.boost_ref_rpm_range, sizeof(boost.boost_ref_rpm_range), engine.rpm, &myrpm_idx);

#define BOOSTGEAR 0
	
	b2 = boost.boostmaxtarget_table[ BOOSTGEAR * IGNADV_SIZE_RPM + myrpm_idx.uindex ];
	rpm_weightu = calc_weightu( engine.rpm, boost.boost_ref_rpm_range, &myrpm_idx );
	weighted = mult16_8 (rpm_weightu, b2) >> 8;

	b1 = boost.boostmaxtarget_table[ BOOSTGEAR * IGNADV_SIZE_RPM + myrpm_idx.lindex ];
	weighted += mult16_8 (0xFFFF - rpm_weightu, b1) >> 8;
	// overflow cannot happen, because of how rpm_weightu is calculated

	engine.boostmax_target = weighted >> 8;
};

void boostgauge_init()
{
	softpwm_init(&boostgauge_pwm, 96); //101kpa
}

//AUDI boostgauge
//Measured equation for dc (0-255) is (MAPbar-0.151)/0.0092 (limit 0-2.5bar)
void boostgauge_act()
{
	uint8_t output;				// DLOW or DHIGH

	uint8_t gauge;
	uint16_t kpa;

	kpa=engine.kpa250; // 0.25kPa

	if(kpa>=43 && kpa<964){
		gauge = mult16_8((kpa-config_ext.boostgauge_offset),config_ext.boostgauge_mul)>>8;
	} else{
		gauge=0;
	}
	//engine.boost_gauge=gauge;
	softpwm_set(&boostgauge_pwm, gauge); //0-255
	if(softpwm_act(&boostgauge_pwm)){
		output = DHIGH;
	} else {
		output = DLOW;
	}
	cli();
	digitalout( config_ext.boostgauge_channel, output ); //AANBOX boostgauge channel
	sei();
};
#endif

// struct softpwm_t boost_pwm_soft;
void boostpwm_act( uint8_t schedule_needed, uint16_t command16, uint8_t outval, uint8_t channel, uint8_t event, uint16_t *stopat_ptr )
{
	if( schedule_needed ) {
		// prepare to schedule event to reset the signal
		uint16_t now;
		now = whats_the_time();

		// command16 meaning changes at this point:
		command16 += now;
		cli();
		// engine.boost_stopat/boostalt_stopat must be changed before digitalout() is called
		// otherwise a racing event could reset output right after
		// we activate the output
		*stopat_ptr = command16;
		sei();
	}

	cli();
	// for inverted output set bit7 of config.boost_channel:
	digitalout( channel, outval );
	sei();

	if( schedule_needed ) {
		cli();
		event_insert( command16, event );
		// without sei_nop_cli(), pending events get priority
		// before other interrupts.
		// sei_nop_cli();
		alarm2core();
		sei();
	}
}

/** actuates the boostalt solenoid;
*  called from mainloop.
* Must be rewritten so boost_
See http://www.vems.hu/wiki/index.php?page=GenBoard%2FUnderDevelopment%2FFirmwareChanges

"simpleboost" bit mean "boostalt=absolute MAP lookup". bit=0 will mean relative lookup
boost_conf.bit1=1 "Simple boost" will be renamed to "MAP absolute lookup". Means that table lookup is according to (RPM,MAP), MAP in the range 0..510 kPa

boost_conf.bit1=0 "PID boost" will be renamed to "relative lookup" which means that table lookup is according to (RPM, MAP - boosttarget). MAP - boosttarget table range is -155 .. +100 kPa. Boosttarget calculated the same way (from same config variables) as the PID boost. You can set TPS scaling to 100% in each bin if you do not use TPS scaling.

Perhaps 2 functions could be used for clarity (but 1 can also be OK).
*/
static void boostcontrol_act_simple(void)
{
	uint16_t command16;
	uint8_t boost, t8;
	uint8_t output;				// DLOW or DHIGH
	uint8_t schedule_needed;	// true if we need to schedule (the other edge)

	command16 = 0;              // just to suppress a gcc warning
	schedule_needed = 0;

	boost = engine.kpa250 >> 3; // current boost level, 2 kpa units
	
	if( boost > config.boost_minpressure ) {
	    uint16_t calc_dc;
	    // engine.boost_ref_dc is calculated in boostref_table_lookup() from boost ref table 
	    calc_dc = engine.boost_ref_dc;

	    if (0 == boost_openloop) {  // closedloop (why called closedloop) ??
			uint8_t target = (boost_data.boost_target16 >> 4);
			// from 12bit to 8bit, 2 kpa units;
			// Kalske says he does NOT want this PWM shutoff with openloop
			if ( ((uint16_t)boost) > ((uint16_t)target) + (BOOST_CONTROL_OVERSHOOT>>1)) {
				// oops 10kPa too high boost, for safety scale down boost DC!
				uint8_t scale;
				scale = boost-target;
				if( scale > 63 ){ // overflow protection
					calc_dc = 0;
				} else {
					scale <<= 2;
					calc_dc = mult16_8(calc_dc, 255-scale) >> 8;
				}
			}

		// Boostsw scale
#define boostsw_enabled     (config_ext.boostsw_conf & 0x08)
		if(boostsw_enabled)
		{
#define boostsw_adc_chan     (config_ext.boostsw_conf & 0x07)
			output = boostsw_adc_chan;
			command16 = mcp3208_data[output];
#ifndef MAINLOOP_TASKER //not needed for tasker
			read_mcp3208(output);
#endif
			if (command16 != 0xFFFF) {
				// data is in range 0..4095
				t8 = command16 >> 4;
				calc_dc = mult16_8(calc_dc, t8) >> 8;
				if (calc_dc >= 255) calc_dc = 255;
				command16 = 0;
			}
		}

		// TPS scale boost_dc
		t8 = boost_tps_downscale();
		calc_dc = mult16_8(calc_dc, t8) >> 8;
		if (calc_dc >= 255) calc_dc = 255;

	    } else {
	    // openloop, adjust DC according to potentiometer reading if configured
            output = boost_openloop_adc_chan;
            if (output < 8) {
          	    command16 = mcp3208_data[output];
#ifndef MAINLOOP_TASKER //not needed for tasker
	            read_mcp3208(output);
#endif
	            if (command16 != 0xFFFF) {
	                // data is in range 0..4095 and we adjust -50..+100%
	                // command = (command * 3/4) / 8 + 128, in range 128..512
	                command16 = (mult16_8(command16, 3) >> 5) + 128;
	                boost_data.dc_adjust = command16 >> 1;
	                calc_dc = mult16_16(calc_dc, command16) >> 8;
	                if (calc_dc >= (255)) calc_dc = (255);
	                command16 = 0;
	            }
	        }
	    }

		if (calc_dc < (BOOSTACT_MINTIME)) {
			// don't activate for very short ontime
			calc_dc = 0;
			output = DLOW;
		} else {
			// range = 0..2047 * BOOSTCONTROL_PWM_PERIOD_8MS * 4usec
			// (BOOSTCONTROL_PWM_PERIOD_8MS * 2048) * 4 usec ON-time
			//command16 = mult16_8(calc_dc, BOOSTCONTROL_PWM_PERIOD_8MS);
			command16 = calc_dc * BOOSTCONTROL_PWM_PERIOD_1MS(config.boost_conf);
			schedule_needed = 1;	// true !!!
			output = DHIGH;
		}
		engine.boostalt_command_duty = (uint8_t)(calc_dc);
	} else {
	/*  Don't activate boost valve below/equal a configurable boost:
		there is no need to wear the valve at cruising
		at low boost where the wastegate is known not to open anyway.
		Without this, the pid15() would raise command, so DHIGH would be likely
	*/
		output = DLOW;

		// be nice on the one decrypting the logs
		engine.boostalt_command_duty = 0;
	}

    boostpwm_act( schedule_needed, command16, output, config_ext.boostalt_channel, EVENT_BOOSTACT + v_shiftup(1), (void*)&engine.boostalt_stopat );
};

/** ================================ 
* boostalt relative and absolute mode
*/
static void boostalt_act_rel_abs(void)
{
	uint16_t command16;
	uint8_t t8;
	uint8_t output; // DLOW or DHIGH
	uint8_t schedule_needed;	// true if we need to schedule (the other edge)
	uint8_t boost_duty;

	if( BOOSTCONF_RELATIVEMODE_CONFIGURED ){ // relative mode
		// relative lookup. boostref table lookup
		// according to MAP - boosttarget. Range = -155 .. +100 kPa
		uint16_t map, btarget;
		map = engine.kpa250; // 1/4 kPa unit
		map += 4*155; // moving to +155 kPa
		btarget = boost_data.boost_target16;
		btarget >>= 1; // going from 1/8 kPa to 1/4 kPa
		if( map > btarget ){
			map -= btarget;
			map >>= 2; // going from 1/4 kPa to 1kPa
			t8 = (map>>8) ? 0xFF : (uint8_t)map;
		} else {
			t8 = 0;
		}
	} else { // absolute mode
		t8 = engine_kpa8();
	}

	boostref_table_lookup(t8);

	// in relative mode we also consider boost minpressure (2 kpa units)
	if( BOOSTCONF_RELATIVEMODE_CONFIGURED
		&& (engine.kpa250 >> 3) <= config.boost_minpressure ){
		boost_duty = 0;
	} else {
	    // engine.boost_ref_dc is calculated in boostref_table_lookup() from boost ref table 
		boost_duty = engine.boost_ref_dc;
	}

	engine.boostalt_command_duty = boost_duty;

	command16 = 0; // suppress a gcc warning
	if (boost_duty == 0){
		schedule_needed = 0;
		output = DLOW;
	} else if( boost_duty == 0xFF){
		schedule_needed = 0;
		output = DHIGH;
	} else {
		schedule_needed = 1;
		output = DHIGH;
		// when boost_duty==FF, we want appr.
		// BOOSTCONTROL_PWM_PERIOD_1MS * FF * 4 usec ON-time
		command16 = boost_duty * BOOSTCONTROL_PWM_PERIOD_1MS(config.boost_conf);
	}

	boostpwm_act( schedule_needed, command16, output, config_ext.boostalt_channel, EVENT_BOOSTACT + v_shiftup(1), (void*)&engine.boostalt_stopat );
};

#ifndef PS2KEYBOARD
static uint16_t my_boost_pid_ref(void)
{
		struct search_table_t myrpm_idx;
/*		uint16_t t1, t2, rpm_weightu, rpm_weightl;
		uint8_t ve_1, ve_2;
		uint8_t boost_ref_dc;

		search_table(ign.adv_table_rpm_range, sizeof(ign.adv_table_rpm_range), engine.rpm, &myrpm_idx);
		rpm_weightu = calc_weightu( engine.rpm, ign.adv_table_rpm_range, &myrpm_idx );
		rpm_weightl = 0xffff - rpm_weightu;

	    ve_11 = boost.boost_ref_table[ IGNADV_SIZE_RPM * engine.gear + myrpm_idx.lindex ];
	    ve_12 = boost.boost_ref_table[ IGNADV_SIZE_RPM * engine.gear + myrpm_idx.uindex ]; 

		t1 = ve_1 * (rpm_weightl >> 8);
		t2 = ve_2 * (rpm_weightu >> 8);
		t2 += t1;
		boost_ref_dc = t2 >> 3;
		boost_ref_dc++; //cheat, interpolation always lose 1
		return boost_ref_dc;*/

	uint8_t ve_11, ve_12;
	uint16_t rpm_weightu, rpm_weightl;
	uint16_t dist_rpm, rpm_vhr;
	uint16_t t1, t2;
	uint8_t rpm_l; // lower RPM


	search_table(ign.adv_table_rpm_range, sizeof(ign.adv_table_rpm_range), engine.rpm, &myrpm_idx);

	// example values are for 5000, 5600, 6000 RPM
	rpm_l = ign.adv_table_rpm_range[myrpm_idx.lindex]; // lower rpm (=50)

	// calculate rpm-distance and column weight:
	dist_rpm = rpm_l * 25;	// (=1250) lower rpm value in 4 RPM resolution (not distance yet)
	rpm_vhr = engine.rpm_vhr;// (=5800) no cli-sei, since rpm_vhr filled from userspace
	rpm_vhr >>= 2;	// (=1400) 4 RPM resolution
	if (rpm_vhr > dist_rpm){ // normally true (rpm higher than rpm_l)
		uint16_t rx21, rx21_d25;
		// dist_rpm = ve.rpm_range[rpm_idx.uindex] - engine.rpm
		dist_rpm = rpm_vhr - dist_rpm;	// (=150) now distance in 4 RPM units
		rx21 = div_65535_x( ign.adv_table_rpm_range[myrpm_idx.uindex] - rpm_l ); // =65535/10=6554
		// rpm_weightl = rx21 / 25 * dist_rpm
		rx21_d25 = div25(rx21); // (=261)
		rpm_weightu = mult16_16(rx21_d25, dist_rpm);
		// 262*250 = 65500 cannot overflow (and 261*150 is very safe)
	} else
		rpm_weightu = 0;

	rpm_weightl = 0xffff - rpm_weightu;

	//separate string for each gear and calc it for pid too
	ve_11 = boost.boost_ref_table[ IGNADV_SIZE_RPM * engine.gear + myrpm_idx.lindex ];
	ve_12 = boost.boost_ref_table[ IGNADV_SIZE_RPM * engine.gear + myrpm_idx.uindex ];
	t1 = ve_11 * (rpm_weightl >> 8);
	t2 = ve_12 * (rpm_weightu >> 8);
	t2 += t1;

	t2>>=3;
	t2++;
	return t2;
}

#endif

static void boostcontrol_act_pid(void)
{
	uint16_t boost, command16;
	uint8_t output;	// DLOW or DHIGH
	uint8_t boost_duty, schedule_needed;	// true if we need to schedule (the other edge)

	command16 = boost_duty = 0;
	// command16 _could_ be used for schedule_needed (before += now is applied)

	/*
		input is currently fixed:
		ADC7 EGO_2
			or
		MAP for MAPtarget

		TODO: make it selectable, including any MCP3208 input
	*/
	if(config.boost_more & _BV(BOOSTTARGET)){
		// boosttarget. Rarely used.
		// Hook up your separate sensor to nernst2, and set amplification to 1x
		boost = sensors11(EGO_2);
	} else {
		// MAPtarget. Commonly used
		boost = engine.kpa250;
	}

	// going to 12 bits [1/8kPa if MAPtarget] :
	boost += boost;

	if( (boost >> 4) > (uint16_t)config.boost_minpressure ){
		uint16_t pid_result;

		pid_set_target( &boostcontrol_pid, boost_data.boost_target16 );
		// with pid15() high-boost will decrease command (more DLOW)
		// which is consistent with DLOW above
		pid_result = pid15_asym( &boostcontrol_pid, boost );

		// TODO: reference position from y[] boost table
		// config.boost_conf bit 3:2 (reference position)
		// configures offset (values: 1/5, 2/5, 3/5, 4/5 of 4096)
		// This slightly effects useful pid_result range (from -2048 .. +2047)
		// so consider this in PID config
#ifdef PS2KEYBOARD
		pid_result += ( 4 + (config.boost_conf & 0x0C)) * 205;
#else//not only gear boost, but also test boostDC table as ref

		//openloop table as ref
		//simple scalar downscaling for refDC from tps dependency
		//scalar shouldn't affect if (tps scale = 100%) for easy tuning at full throttle first
		//tps scale = any%, and scalar = 100%, refDC = tps scale
		//tps scale = any%, and scalar = 0%, refDC = 100%
		//tps scale = 100%, and scalar = any, refDC = 100%
		//tps scale = 70%, and scalar = 50%, refDC = 85%
		//tps scale = 70%, and scalar = 100%, refDC = 70%
		//tps scale = 70%, and scalar = 200%, refDC = 40%
		//tps scale = 50%, and scalar = 70%, refDC = 65%
		//tps scale = 50%, and scalar = 50%, refDC = 75%
		if (boost_PWM_gear_depending) {    // config bit for using boost ref depending on rpm/gear, like on 1.1.47 http://www.vems.hu/wiki/index.php?page=MembersPage%2FDamirMuha%2FBoostDependingGear
		   uint16_t t16;

		   t16 = config_ext.boost_tps_dc_scale * (0xFF-boost_tps_downscale());
		   if (t16&0x8000) // we cant substract more than 100%
		   {
			   t16 = 0;// refDC = 0
		   }
		   else
		   {
			   t16 = 0x7FFF-t16; // = 0 for 100%
			   t16 = mult16_16(t16, my_boost_pid_ref())>>16;
		   }
		   pid_result += t16;	// engine.boost_ref_dc<<1 = 100% = 4096
		}
		else
		{
		pid_result += ( 4 + (config.boost_conf & 0x0C)) * 205;      //    simple ref boost position
		}

		
#endif
		// prevent underflow, overflow.
		// 3 outputs possible: always low, always high
		// and something between
		if( pid_result >= 0x8000 || pid_result < BOOSTACT_MINTIME){
			// always OFF
			schedule_needed = 0;
			// boost_duty = 0; initialized before
			output = DLOW;
		} else if( pid_result >= 4096 ){ // compare against (4096 - BOOSTACT_MINTIME) ?
			// fully ON! good luck Mr. Gorsky!
			schedule_needed = 0;
			output = DHIGH;
			boost_duty = 0xFF;
		} else {
			schedule_needed = 1;	// true !!!
			output = DHIGH;
			boost_duty = (uint8_t)(pid_result >> 4);
		// when pid_result is 4095, we want appr.
		// (BOOSTCONTROL_PWM_PERIOD_8MS * 2048) * 4 usec ON-time
			command16 = mult16_8( pid_result, BOOSTCONTROL_PWM_PERIOD_8MS);
			command16 >>= 1;
		}

	} else {
	/*
		Don't activate boost valve below/equal a configurable boost:
		there is no need to wear the valve at cruising
		at low boost where the wastegate is known not to open anyway.
		Without this, the pid15() would raise command, so DHIGH would be likely
	*/
		schedule_needed = 0;
		// boost_duty = 0; initialized before
		output = DLOW;
	}
	engine.boost_command_duty = boost_duty;
    boostpwm_act( schedule_needed, command16, output, config.boost_channel, EVENT_BOOSTACT + 0, (void*)&engine.boost_stopat );
}

void boostcontrol_init()
{
        pid_init(&boostcontrol_pid, (struct pid_parameters_t*) &config.boost_pid_kp);
}

/**
*	returns the calculated fully-mappable boost-target in 12..13 bits:
*	16 * ( config.boost_targetoffs + b[RPM] * t[TPS] / 256)
*
*	Note0: currently only TPS and RPM-dependent (gears: later).
*
*	8 bits should be enough for engine.boostmax_target:
*	eg. 2kPa unit for max 512kPa boost
*
*	Note1:
*		unit is [1/8kPa if MAPtarget] 
*		unit is fully user definable (if boost-target)
*
*	Note2:
*		user chooses with config.boost_targetoffs
*	if she wants to reference
*	configured value from 100kPa or 0kPa
*	It matters, since it is applied after
*	engine.boostmax_target is downscaled according to the tpstransfer...
*
*	For MAP-target, she probably wants 0kPa,
*	while for boost-target she probably wants 100kPa
*/

static uint16_t boost_target(void)
{
	uint16_t egt16, boost_target; // [1/8 kPa] with MAPtarget
	uint8_t t8, threshold;

	// set TPStransfer table to all FF for no TPS downscale
	boost_target = engine.boostmax_target;
 	if( boostsimple_configured && (0 == boost_openloop) ){
		boost_target <<= 4; // NO TPS downscale
	} else {
		boost_target = mult16_8( boost_target, boost_tps_downscale()) >> 4;
	}
	// originally it was finely adjustable config.boost_targetoffs :
	boost_target += (config.boost_more & _BV(BOOST_TARGETOFFS)) ? 800 : 0; // 100 or 0 kPa

#ifndef PS2KEYBOARD
	if (engine.gear<5) //not 6 gear
	{//apply downscale for target
		boost_target = mult16_8(boost_target, config_ext.boost_mult[engine.gear])>>7;
	}
#endif

	if( (t8=engine.mat) > (threshold = config.boost_matdec) ){
		t8 -= threshold;
		uint16_t cut16 = t8;
		cut16 <<= 5 ; // +10C => 320 [1/8 kPa] = -40 kPa
		boost_target = boost_target > cut16 ? boost_target-cut16 : 0;
	}

	egt16 = engine.egt;
	t8 = egt16 >> 8; // rail above 2040C, not overflow (never happen)
	t8 = t8 > 7 ? 255 : egt16>>3; // going to [8C]
	if( t8 > (threshold = config.boost_egtdec) ){
		t8 -= threshold;
		uint16_t cut16 = t8 * 64 ; // +40C = +5 => 320 * [1/8 kPa) = -40 kPa
		boost_target = boost_target > cut16 ? boost_target-cut16 : 0;
	}

	return boost_target;
}

/**
* void wrapper for boostcontrol functions
*/
void ml_variable_boost(void)
{
	boostrpm_lookup(); // sets engine.boostmax_target (used immediately : )
	boost_data.boost_target16 = boost_target();
	if( ! digitalout_disabled( config_ext.boostalt_channel ) ){
		uint8_t t8 = engine_kpa8(); // absolute lookup in modes 01, 10, 11

		if( boostsimple_configured ){ // mode 10 is closedloop and 11 openloop:
			boostref_table_lookup(t8);
			boostcontrol_act_simple();
		} else {
			boostalt_act_rel_abs();
		}
	}
/* }
* will be split to separate function if we want different PWM period:
* void ml_variable_boostalt(void)
* {
*/
	if( ! digitalout_disabled( config.boost_channel ) ){
		boostcontrol_act_pid();
	}
}
#endif	// BOOSTCONTROL
